#include <stdio.h>

typedef struct treepairnode treepairnode;

typedef struct {
	int intero;
	char *stringa;
	treepairnode *i, *s;
} datanode;

struct treepairnode {
	datanode *data;
	treepairnode *left, *right, *up;
};

typedef struct {
	treepairnode *iroot, *sroot;
} treepair;

/* crea un treepair vuoto */
treepair *createtreepair(void);

/* visita inordine il searchtree (di interi o di stringhe) la cui radice e' p */
/* e applica a ogni nodo la funzione op */
void inorder(treepairnode *p, void (*op)(treepairnode *));

/* restituisce la radice del searchtree di interi nel treepair di radice p */
/* NULL se p e' NULL */
treepairnode *getintroot(treepair *p);

/* restituisce la radice del searchtree di stringhe nel treepair di radice p */
/* NULL se p e' NULL */
treepairnode *getstrroot(treepair *p);

/* restituisce il valore intero associato al nodo puntato da p */
/* -1 se p e' NULL o p non punta ad alcun datanode */
int getintdatum(treepairnode *p);

/* restituisce la stringa associata al nodo puntato da p */
/* NULL se p e' NULL o p non punta ad alcun datanode */
char *getstrdatum(treepairnode *p);

/* restituisce il puntatore a un nodo nel treeepair di radice p */
/* che abbia valore intero associato uguale a k */
/* NULL se p e' NULL */
treepairnode *intsearch(treepair *p, int k);

/* restituisce il puntatore a un nodo nel treeepair di radice p */
/* che abbia la stringa associata uguale a k */
/* NULL se p e' NULL */
treepairnode *strsearch(treepair *p, char *k);

/* restituisce il minimo del searchtree (di interi o di stringhe) di radice p */
treepairnode *treemin(treepairnode *p); /* si assume q != NULL */

/* restituisce il minimo del searchtree (di interi o di stringhe) di radice p */
treepairnode *treemax(treepairnode *p); /* si assume q != NULL */

/* restituisce il successore di q nel searchtree cui q appartiene */
treepairnode *treesucc(treepairnode *q); /* si assume q != NULL */

/* restituisce il predecessore di q nel searchtree cui q appartiene */
treepairnode *treepred(treepairnode *q); /* si assume q != NULL */

/* inserisce nel treepair un datanode con valore intero k e stringa s */
void insert(treepair *root, int k, char *s);

/* cancella dal treepair il nodo puntato da q, il datanode associato e il */
/* nodo associato al datanode nell'altro searchtree del treepair */
void delete(treepair *root, treepairnode *q); /* si assume q != NULL */





