/* zaino con oggetti irripetibili */
/* Nota: sono commentate solo le differenze rispetto a zaino.c */
#include <stdio.h>
#include <stdlib.h>

#define DIMWORD 20

#define max(a,b)		((a)>(b)?(a):(b))
#define posto(r,c,n)		((r)*(n)+(c))

void leggidati(FILE *fp, char *nome, int *peso, int *valore, int n)
{
	int i;

	for(i = 0; i < n; i++)
		fscanf(fp,"%s %d %d",nome + (DIMWORD * i), peso + i, valore + i);
}

/* calcola una soluzione ottimale al problema dello zaino senza ripetizione di oggetti */
/* zaino e' una matrice di (limitepeso + 1)*(n + 1) interi tale che */
/* zaino[j][i] conterra' il valore di una soluzione ottima al problema di capacita' j */
/* definito sull'insieme dei primi i oggetti */
int *calcolazaino(int *peso, int *valore, int n, int limitepeso)
{
	int *zaino;
	int i,j,k,v1,v2;

	/* alloca la matrice zaino */
	if(!(zaino = calloc((limitepeso + 1) * (n + 1), sizeof(int)))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-3);
	}
	/* ciclo sugli oggetti */
	for(i = 1;i <= n;i++)
		/* ciclo sulle capacita' */
		for(j = 0;j <= limitepeso;j++) {
			/* assegna a v1 il valore della soluzione ottimale di capacita j sui primi i-1 oggetti */			
			v1 = zaino[posto(j,i-1,n+1)];
			/* determina la capacita' di uno zaino ottenuto togliendo l'i-esimo oggetto da uno zaino di capacita' j */
			/* nota che il peso dell'i-esimo oggetto e' in peso[i-1] */
			k = j - peso[i-1];
			/* assegna a v2 il valore della soluzione ottenuta aggiungendo */
			/* l'i-esimo oggetto */
                  	/* alla soluzione ottimale per capacita' k sui primi i-1 oggetti. */
			/* Ovviamente deve valere k >= 0 */
			v2 = k < 0 ? 0 : zaino[posto(k,i-1,n+1)] + valore[i-1];
			/* assegna a zaino[k] la migliore delle due soluzioni */
			zaino[posto(j,i,n+1)] = max(v1,v2);
		}
	return zaino;
}

/* stampa la soluzione ottima contenuta nella matrice zaino */
void stampazaino(int *zaino, int *peso, char *nome, int n, int limitepeso)
{
	int i,j;
	
	/* il valore della soluzione e' in zaino[limitepeso][n] */
	printf("Il valore totale dello zaino e' %d\n",zaino[posto(limitepeso,n,n+1)]);
	printf("Lo zaino contiene: \n");

	/* ciclo per determinare gli oggetti contenuti */
	for(i = limitepeso, j = n; j; j--)
		/* se zaino[i,j] != zaino[i,j-1], allora l'ultimo oggetto aggiunto */
		/* allo zaino per ottenere zaino[i,j] e' stato il j-esimo oggetto */
		if(zaino[posto(i,j,n+1)] != zaino[posto(i,j-1,n+1)]) {
			printf("%s\n", nome + (j-1) * DIMWORD);
			/* aggiorno l'indice i, sottraendogli il peso del j-esimo oggetto */
			i -= peso[j-1];
		}	
	putchar('\n');
}

int main(int ac, char **av)
{
	FILE *fp;
	int n, limitepeso;
	int *peso, *valore, *zaino;
	char *nome;
	

	if(ac != 2) {
		fprintf(stderr, "Uso: %s <file>\n", av[0]);
		exit(-1);
	}
	if(!(fp = fopen(av[1],"r"))) {
		fprintf(stderr, "Errore: non riesco ad aprire il file %s\n", av[1]);
		exit(-2);
	}

	fscanf(fp,"%d%d",&limitepeso,&n);
	peso = calloc(n,sizeof(int));
	valore = calloc(n,sizeof(int));
	nome = calloc(n, DIMWORD + 1);
	
	if(!peso || !valore || !nome) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-3);
	}

	leggidati(fp,nome,peso,valore,n);
	zaino = calcolazaino(peso,valore,n,limitepeso);
	stampazaino(zaino,peso,nome,n,limitepeso);

	return 0;
}
